import copy
import json
import logging
import os


class WatchedFile:
    """This class tracks a file and updates the in-memory contents when the file is modified."""

    def __init__(self, file_path, log):
        self._cached_content = None
        self._last_mtime_ns = None
        self.file_path = file_path
        self._log = log

    def read(self):
        if self.requires_read():
            self._read_modified()
        else:
            self._log.info(
                f"Returning content of [{self.file_path}] updated at time {self._last_mtime_ns}"
            )

        # the file_content should have been read by this point
        assert self._cached_content is not None
        return copy.copy(self._cached_content)

    def _update_cache(self, updated_content):
        self._cached_content = updated_content
        self._last_mtime_ns = os.stat(self.file_path).st_mtime_ns
        self._log.info(f"Read [{self.file_path}] from disk at {self._last_mtime_ns}")

    def _read_modified(self):
        with open(self.file_path, "r") as f:
            self._update_cache(f.read())

    def requires_read(self):
        try:
            return (
                self._cached_content is None
                # this part is in the second branch of the `or` so that if
                # the file does not exist yet, there is still time to create it
                or (os.stat(self.file_path).st_mtime_ns != self._last_mtime_ns)
            )
        except:
            return False


class WatchedJsonFile(WatchedFile):
    def _read_modified(self):
        try:
            with open(self.file_path, "r") as f:
                self._update_cache(json.load(f))
        except:
            logging.error(f"Unable to read modified file. Returning empty value...")
            self._update_cache({})

    def get_key(self, key, default_value=None):
        """
        Tries to read and parse the json file and get the key from root.

        Provided default_value is used if key found is not found.
        """

        try:
            value = self.read().get(key, default_value)
            return value
        except FileNotFoundError as error:
            logging.error(
                f"The watched file({self.file_path}) does not exist to get key: {key}"
            )
            return default_value



import logging
import os
from typing import Optional

import botocore
from aiobotocore.session import AioSession, get_session
from traitlets.config import SingletonConfigurable

from amzn_sagemaker_aiops_jupyterlab_extension.exceptions import S3ObjectNotFoundError
from amzn_sagemaker_aiops_jupyterlab_extension.utils.app_metadata import (
    get_domain_id,
    get_partition,
    get_region_name,
    get_space_name,
    get_stage,
)

logging.basicConfig(level=os.environ.get("LOGLEVEL", "INFO"))

# Values currently set to match LooseLeafNb2Kg
MAX_RETRY_ATEMPTS = 4
CONNECTION_TIMEOUT = 60  # seconds
READ_TIMEOUT = 60  # seconds

LOOSELEAF_STAGE_MAPPING = {"devo": "beta", "loadtest": "gamma"}


class BaseAysncBoto3Client:
    cfg: any
    region_name: str
    partition: str
    sess: AioSession

    def __init__(self, region_name: str, partition: str, model_data_path=None):
        self.cfg = botocore.client.Config(
            connect_timeout=CONNECTION_TIMEOUT,
            read_timeout=READ_TIMEOUT,
            retries={"max_attempts": MAX_RETRY_ATEMPTS},
        )
        self.region_name = region_name
        self.partition = partition
        # This allows us to load any additional service model files. That could allows us to have private service clients or desired methods.
        # https://botocore.amazonaws.com/v1/documentation/api/latest/reference/loaders.html
        # This does not imapct any other clients created or running in environemnt,
        if model_data_path is not None:
            self.sess = get_session(
                env_vars={"data_path": (None, None, model_data_path, None)}
            )
        else:
            self.sess = get_session()


class S3AsyncBoto3Client(BaseAysncBoto3Client, SingletonConfigurable):
    """This class is a Singleton that provides a simple async wrapper on s3 boto3 client.
    This wrapper class uses aiobotocore, a library that provides async client to aws services.
    """

    def __init__(self, region_name, partition_name, model_data_path=None):
        super().__init__(region_name, partition_name, model_data_path)

    def _create_s3_client(self):
        create_client_args = {
            "service_name": "s3",
            "config": self.cfg,
            "region_name": self.region_name,
        }
        return self.sess.create_client(**create_client_args)

    async def get_object(
        self, bucket: str, key: str, expected_bucket_owner: Optional[str] = None
    ) -> bytes:
        try:
            async with self._create_s3_client() as s3_client:
                if expected_bucket_owner is not None:
                    response = await s3_client.get_object(
                        Bucket=bucket,
                        Key=key,
                        ExpectedBucketOwner=expected_bucket_owner,
                    )
                else:
                    response = await s3_client.get_object(Bucket=bucket, Key=key)
                async with response["Body"] as stream:
                    content = await stream.read()
        except botocore.exceptions.ClientError as error:
            logging.error(f"Failed to get s3 object. {error.response}")
            raise
        return content

    async def head_object(
        self, bucket: str, key: str, expected_bucket_owner: Optional[str] = None
    ) -> dict:
        try:
            async with self._create_s3_client() as s3_client:
                if expected_bucket_owner is not None:
                    response = await s3_client.head_object(
                        Bucket=bucket,
                        Key=key,
                        ExpectedBucketOwner=expected_bucket_owner,
                    )
                else:
                    response = await s3_client.head_object(Bucket=bucket, Key=key)
        except botocore.exceptions.ClientError as error:
            if error.response["Error"]["Code"] == "404":
                raise S3ObjectNotFoundError(
                    f"Object does not exist. bucket: {bucket}, key: {key}"
                )
            else:
                raise
        return response


class SageMakerAsyncBoto3Client(BaseAysncBoto3Client, SingletonConfigurable):
    """
    A Singleton Class that provides a simple async wrapper on sagemaker boto3 client.
    This wrapper class uses aiobotocore, a library that provides async client to aws services.
    """

    def __init__(
        self,
        region_name: str,
        partition_name: str,
        model_data_path: Optional[str] = None,
    ):
        super().__init__(region_name, partition_name, model_data_path)
        self.default_domain_id = get_domain_id()
        self.default_space_name = get_space_name()

    def _create_sagemaker_client(self):
        stage = get_stage()

        create_client_args = {
            "service_name": "sagemaker",
            "config": self.cfg,
            "region_name": self.region_name,
        }
        if stage is not None and stage != "" and stage.lower() != "prod":
            endpoint_stage = LOOSELEAF_STAGE_MAPPING[stage.lower()]
            create_client_args["endpoint_url"] = (
                f"https://sagemaker.{endpoint_stage}.{self.region_name}.ml-platform.aws.a2z.com"
            )

        return self.sess.create_client(**create_client_args)

    async def describe_domain(self, domain_id=None):
        try:
            domainId = self.default_domain_id if domain_id is None else domain_id
            logging.info(f"Received request to describe domain with Id - ({domainId})")
            async with self._create_sagemaker_client() as sm_client:
                response = await sm_client.describe_domain(DomainId=domainId)
                logging.info(f"Successfuly described domain with Id: ({domainId})")
        except botocore.exceptions.ClientError as error:
            logging.error(
                f"Failed to describe domain for domainId ({domainId}). {error.response}"
            )
            raise error
        return response

    async def describe_user_profile(self, domain_id=None, user_profile_name=None):
        try:
            domainId = self.default_domain_id if domain_id is None else domain_id
            logging.info(
                f"Received request to describe user-profile with DomainId ({domainId}) and profile-name ({user_profile_name}) "
            )
            async with self._create_sagemaker_client() as sm_client:
                response = await sm_client.describe_user_profile(
                    DomainId=domainId, UserProfileName=user_profile_name
                )
                logging.info(
                    f"Successfuly described user-profile ({user_profile_name}) in domain with Id ({domainId})"
                )
        except botocore.exceptions.ClientError as error:
            logging.error(
                f"Failed to describe user-profile ({user_profile_name}) in domain ({domainId}). {error.response}"
            )
            raise error
        return response

    async def describe_space(self, domain_id=None, space_name=None):
        try:
            domainId = self.default_domain_id if domain_id is None else domain_id
            spaceName = self.default_space_name if space_name is None else space_name
            logging.info(
                f"Received request to describe space with DomainId ({domainId}) and space-name ({spaceName}) "
            )
            async with self._create_sagemaker_client() as sm_client:
                response = await sm_client.describe_space(
                    DomainId=domainId, SpaceName=spaceName
                )
                logging.info(
                    f"Successfuly described space ({spaceName}) in domain with Id ({domainId})"
                )
        except botocore.exceptions.ClientError as error:
            logging.error(
                f"Failed to describe space ({spaceName}) in domain ({domainId}). {error.response}"
            )
            raise error
        return response

    async def list_projects(self):
        projects = []
        next_token = None
        try:
            logging.info("Received request to list projects")
            async with self._create_sagemaker_client() as sm_client:
                while True:
                    params = {}
                    if next_token:
                        params["NextToken"] = next_token
                    response = await sm_client.list_projects(**params)
                    projects.extend(response.get("ProjectSummaryList", []))
                    next_token = response.get("NextToken")
                    if not next_token:
                        break
                logging.info("Successfuly listed projects")
        except botocore.exceptions.ClientError as error:
            logging.error("Failed to list projects")
            raise error
        return {"ProjectSummaryList": projects}


"""
    Returns the Sagemaker Client.
    create_new: boolean flag that allows to force-create a new instance for sagemaker client. Default is False
"""


def get_sagemaker_client(create_new=False):
    PACKAGE_ROOT = os.path.abspath(os.path.dirname(__file__))
    model_path = os.path.join(PACKAGE_ROOT, "botocore_model")

    if create_new:
        return SageMakerAsyncBoto3Client(get_region_name(), get_partition(), model_path)
    else:
        return SageMakerAsyncBoto3Client.instance(
            get_region_name(), get_partition(), model_path
        )


"""
    Returns the S3 Client.
    create_new: boolean flag that allows to force-create a new instance for s3 client. Default is False
"""


def get_s3_client(create_new=False):
    if create_new:
        return S3AsyncBoto3Client(get_region_name(), get_partition())
    else:
        return S3AsyncBoto3Client.instance(get_region_name(), get_partition())



import os
import tarfile
from pathlib import Path
from typing import Optional

from async_lru import alru_cache

from amzn_sagemaker_aiops_jupyterlab_extension.clients import get_s3_client
from amzn_sagemaker_aiops_jupyterlab_extension.exceptions import NotebookTooLargeError
from amzn_sagemaker_aiops_jupyterlab_extension.utils.app_metadata import (
    get_aws_account_id,
    get_region_name,
)


def _find_output_notebook(extract_path):
    """
    Find the notebook file that ends with -output.ipynb in the extracted directory
    """
    try:
        # Search recursively for files ending with -output.ipynb
        notebook_files = list(Path(extract_path).rglob("*-output.ipynb"))

        if notebook_files:
            # Return the first matching notebook file
            return str(notebook_files[0])
        return None
    except Exception:
        return None


def _extract_tar_gz(tar_path, extract_path):
    """
    Safely extract tar.gz file to the specified path
    """
    try:
        with tarfile.open(tar_path, "r:gz") as tar:
            # Check for any files that would extract outside the target directory
            for member in tar.getmembers():
                member_path = os.path.join(extract_path, member.name)
                if not os.path.abspath(member_path).startswith(
                    os.path.abspath(extract_path)
                ):
                    raise Exception("Attempted path traversal in tar file")

            # Extract the files
            tar.extractall(path=extract_path)
        return True
    except Exception:
        raise


@alru_cache(maxsize=5, ttl=60)
async def _get_object_size(bucket: str, key: str) -> Optional[int]:
    account_id = get_aws_account_id()
    s3_client = get_s3_client()
    response = await s3_client.head_object(bucket, key, account_id)
    return response.get("ContentLength")


@alru_cache(maxsize=5, ttl=60)
async def _get_notebook_content(key: str):
    print(f"\nStarting _get_notebook_content with key: {key}")
    # Construct the bucket name using the correct SageMaker format
    account_id = get_aws_account_id()
    region = get_region_name()
    bucket = f"sagemaker-{region}-{account_id}"

    object_size = await _get_object_size(bucket, key)

    if not object_size or object_size > 26214400:
        raise NotebookTooLargeError(f"Object {key} is too large to fit into memory")

    response = await get_s3_client().get_object(bucket, key, account_id)

    return response



import json
from os import path
from pathlib import Path
from .handlers import register_handlers
from aws_embedded_metrics.config import get_config

try:
    from ._version import __version__
except ImportError:
    __version__ = "dev"

HERE = Path(__file__).parent.resolve()

# Handle missing labextension directory gracefully
try:
    with (HERE / "labextension" / "package.json").open(encoding="utf-8") as fid:
        data = json.load(fid)
except (FileNotFoundError, json.JSONDecodeError):
    data = {"name": "@amzn/amzn_sagemaker_aiops_jupyterlab_extension"}


# Path to the frontend JupyterLab extension assets
def _jupyter_labextension_paths():
    return [{"src": "labextension", "dest": data["name"]}]


def _jupyter_server_extension_points():
    return [
        {
            "module": "amzn_sagemaker_aiops_jupyterlab_extension",
        }
    ]


# Entrypoint of the server extension
def _load_jupyter_server_extension(nb_app):
    nb_app.log.info(f"Loading SageMaker JupyterLab server extension {__version__}")

    # configure EMF logger
    emf_config = get_config()
    emf_config.namespace = "StudioAIOpsJupyterLabExtensionServer"

    register_handlers(nb_app)


load_jupyter_server_extension = _load_jupyter_server_extension



# Standard library imports
import json
import os
from pathlib import Path

# Third-party imports
import tornado
from jupyter_server.base.handlers import JupyterHandler
from jupyter_server.utils import url_path_join

from amzn_sagemaker_aiops_jupyterlab_extension.utils.pipelines_output_util import (
    _extract_tar_gz,
    _find_output_notebook,
    _get_notebook_content,
)

# Constants
DEFAULT_HOME_DIRECTORY = "/home/sagemaker-user"
HOME_PATH = os.environ.get("HOME", DEFAULT_HOME_DIRECTORY)


def build_url(web_app, endpoint):
    base_url = web_app.settings["base_url"]
    # Remove any trailing slashes
    base_url = base_url.rstrip("/")
    endpoint = endpoint.lstrip("/")
    return url_path_join(base_url, endpoint)


def register_handlers(nbapp):
    web_app = nbapp.web_app
    host_pattern = ".*$"

    # Add debug logging
    nbapp.log.info("Registering handlers for amzn_sagemaker_aiops_jupyterlab_extension")
    nbapp.log.info(f"Base URL: {web_app.settings['base_url']}")

    handlers = [
        (
            build_url(web_app, "/aws/sagemaker/api/pipelines/get-output"),
            PipelinesOutputHandler,
        ),
    ]
    web_app.add_handlers(host_pattern, handlers)


class PipelinesOutputHandler(JupyterHandler):
    """Handler to download and extract output.tar.gz file from user's SageMaker S3 bucket"""

    @tornado.web.authenticated
    async def get(self):
        self.set_header("Content-Type", "application/json")
        try:
            # Get required parameters from query parameters
            pipeline_name = self.get_argument("pipelineName", None)
            step_name = self.get_argument("stepName", None)
            job_name = self.get_argument("jobName", None)

            # Validate required parameters
            if not all([pipeline_name, step_name, job_name]):
                self.set_status(400)
                self.finish(
                    json.dumps(
                        {
                            "error": "Missing required parameters",
                            "details": "pipelineName, stepName, and jobName are required",
                        }
                    )
                )
                return

            # Construct the S3 key path
            s3_key = f"sagemaker-pipelines/{pipeline_name}/{step_name}/output/{job_name}/output/output.tar.gz"

            # Create the target directory path using HOME_PATH from environment
            home_path = os.environ.get("HOME", "/tmp")  # Use /tmp as fallback
            target_dir = Path(home_path) / pipeline_name / step_name
            target_file = target_dir / "output.tar.gz"

            try:
                # Create directories if they don't exist
                target_dir.mkdir(parents=True, exist_ok=True)

                # Download the file from S3
                self.log.info(f"Downloading from: {s3_key}")
                self.log.info(f"Saving to: {target_file}")

                file_content = await _get_notebook_content(s3_key)

                # Save the file to the target location
                with open(target_file, "wb") as f:
                    f.write(file_content)

                # Extract the tar.gz file
                self.log.info(f"Extracting {target_file} to {target_dir}")
                _extract_tar_gz(str(target_file), str(target_dir))

                # Find the output notebook
                notebook_path = _find_output_notebook(target_dir)

                # Remove the tar.gz file after extraction
                if target_file.exists():
                    target_file.unlink()

                response_data = {
                    "success": True,
                    "message": "File downloaded and extracted successfully",
                    "extractedPath": str(target_dir),
                }

                # Add notebook path to response if found
                if notebook_path:
                    response_data["notebookPath"] = notebook_path

                self.set_status(200)
                self.finish(json.dumps(response_data))

            except Exception as e:
                self.log.error(f"Error processing file from S3: {str(e)}")
                # Clean up the downloaded file if it exists
                if target_file.exists():
                    try:
                        target_file.unlink()
                    except Exception as cleanup_error:
                        self.log.error(f"Error cleaning up file: {str(cleanup_error)}")

                self.set_status(500)
                self.finish(
                    json.dumps(
                        {
                            "error": "Error processing file",
                            "details": str(e),
                        }
                    )
                )

        except Exception as e:
            self.log.error(f"Error in PipelinesOutputHandler: {str(e)}")
            self.set_status(500)
            self.finish(
                json.dumps({"error": "Internal server error", "details": str(e)})
            )



class S3ObjectNotFoundError(Exception):
    """Exception throw if object doesn't not exist"""

    pass


class NotebookTooLargeError(Exception):
    """Exception throw if notebook size larger than NOTEBOOK_SIZE_LIMIT_IN_BYTES"""

    pass


class DownloadDirectoryNotFoundError(Exception):
    """Exception throw if download directory is not found"""

    pass
